use dotenv;
use kuchiki;
use kuchiki::traits::TendrilSink;
use lettre::transport::smtp::authentication::Credentials;
use lettre::{Message, SmtpTransport, Transport};
use reqwest;
use reqwest::blocking::Response;
use reqwest::redirect::Policy;
use reqwest::Error;
use sentry::ClientInitGuard;
use serde::{Deserialize, Serialize};
use std::env;
use std::sync::mpsc::channel;
use std::sync::mpsc::Sender;
use std::thread;
use std::time;
use threadpool;

#[derive(Clone, Debug)]
struct OneclickAuth {
    base_url: String,
    client_url: String,
    client_number: String,
    username: String,
    password: String,
    organization: String,
    parent_organization: String,
    token: String,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
struct UserInfo {
    sub: String,
    #[serde(rename = "http://claims.addison.de/organizationguid")]
    organization: String,
    #[serde(rename = "http://claims.addison.de/parentorganizationguid")]
    parent_organization: String,
}

impl OneclickAuth {
    fn to_documents(&self) -> String {
        format!(
            "{}/oneclick/api/content/organizations/{}/documents",
            self.base_url, self.organization
        )
    }

    fn to_document_download(&self, blob_id: &String) -> String {
        format!(
            "{}/oneclick/api/content/organizations/{}/downloads/blobs/{}",
            self.base_url, self.parent_organization, blob_id
        )
    }

    fn to_document_put_state(&self, blob_id: &String) -> String {
        format!(
            "{}/oneclick/api/content/organizations/{}/documents/{}/state",
            self.base_url, self.organization, blob_id
        )
    }

    fn to_user_info_url(&self) -> String {
        // GET https://services.portal-bereich.de/servicehosts/authority/connect/userinfo
        format!("{}/authority/connect/userinfo", self.base_url)
    }

    // we need to obtain a token - with implicit flow -.-
    async fn obtain_token(&mut self) {
        let first_url = format!("{}/authority/connect/authorize", self.base_url);
        let second_url = format!("{}/authority/Account/Login", self.base_url);
        let policy = Policy::custom(|attempt| {
            if attempt.previous().len() > 5 {
                attempt.error("too many redirects")
            } else if attempt.url().path() == "/html/" {
                // prevent redirects to 'callback url'
                attempt.stop()
            } else {
                attempt.follow()
            }
        });
        let client = reqwest::Client::builder()
            .cookie_store(true)
            /*.timeout(time::Duration::from_secs(30))
            .connect_timeout(time::Duration::from_secs(30))*/
            .redirect(policy)
            .build()
            .expect("Client");

        let redirect_uri = format!("{}/html/", self.client_url);
        let resp = client
            .get(&first_url)
            .query(&[
                ("client_id", "implicitportal"),
                ("redirect_uri", redirect_uri.as_str()),
                ("response_type", "id_token token"),
                ("scope", "oneclick openid profile"),
                ("state", "I2hvbWU="),
                ("nonce", "0.23324117276355705"),
            ])
            .header(
                "User-Agent",
                "Mozilla/5.0 (X11; Linux x86_64; rv:95.0) Gecko/20100101 Firefox/95.0",
            )
            .header("Referer", self.client_url.clone())
            .send()
            .await
            .expect("Login page");

        let html = resp.text().await.expect("HTML");
        let document = kuchiki::parse_html().one(html);
        let x = document.select("input[type=hidden]").expect("HTML form");
        let mut oiaf = OpenIdAuthForm {
            return_url: String::from(""),
            uri: String::from(""),
            client_number: self.client_number.clone(),
            username: self.username.clone(),
            password: self.password.clone(),
            remember_login: String::from("false"),
            button: String::from("login"),
            request_verification_token: String::from(""),
        };
        for element in x {
            let attrs = element.attributes.borrow();
            let node_name = attrs.get("name").unwrap();
            let node_value = attrs.get("value").unwrap();
            if node_name == "ReturnUrl" {
                oiaf.return_url = String::from(node_value);
            }
            if node_name == "Uri" {
                oiaf.uri = String::from(node_value);
            }
            if node_name == "__RequestVerificationToken" {
                oiaf.request_verification_token = String::from(node_value);
            }
        }
        // Send request
        let client_referer = self.client_url.clone();
        let resp = client.post(&second_url)
            .query(&[
                ("ReturnUrl", oiaf.return_url.clone())
            ])
            .form(&oiaf)
            .header("User-Agent", "Mozilla/5.0 (X11; Linux x86_64; rv:95.0) Gecko/20100101 Firefox/95.0")
            .header("Referer", client_referer.as_str())
            .header("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8")
            .send()
            .await
            .expect("Login successful");

        let headers = resp.headers();
        let redirect_url_hdr = headers.get("Location").expect("Location");
        let redirect_url =
            reqwest::Url::parse(redirect_url_hdr.to_str().expect("String")).expect("URL");
        let fragment = redirect_url.fragment().expect("Fragment");
        let fparse = form_urlencoded::parse(fragment.as_bytes());
        let mut token = String::from("");
        for element in fparse {
            let key = element.0.to_string();
            let value = element.1.to_string();
            if key == String::from("access_token") {
                token.push_str(value.as_str());
            }
        }
        self.token = token;
    }

    async fn get_user_info(&mut self) {
        let client = reqwest::Client::new();
        let url = self.to_user_info_url();
        let resp = client
            .get(&url)
            .bearer_auth(&self.token)
            .header(
                "User-Agent",
                "Mozilla/5.0 (X11; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0",
            )
            .header("Accept", "application/json, text/javascript, */*; q=0.01")
            .send()
            .await
            .expect("Claims...");

        let claims: UserInfo = resp.json().await.expect("Claims...");
        self.organization = claims.organization.clone();
        self.parent_organization = claims.parent_organization.clone();
    }
}

#[derive(Clone, Debug)]
struct SeafileApi {
    base_url: String,
    password: String,
    token: String,
    repo_path: String,
    upload_link: String,
}

#[derive(Clone, Debug, Serialize)]
struct SeafileApiAuth {
    username: String,
    password: String,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
struct OpenIdAuthForm {
    return_url: String,
    uri: String,
    client_number: String,
    username: String,
    password: String,
    #[serde(rename = "button")]
    button: String,
    remember_login: String,
    #[serde(rename = "__RequestVerificationToken")]
    request_verification_token: String,
}

impl SeafileApi {
    fn to_upload_link_url(&self) -> String {
        format!("{}/via-repo-token/upload-link", self.base_url)
    }

    async fn auth(&mut self) {
        self.token = format!("Token {}", self.password.clone());
    }

    async fn get_upload_link(&mut self) {
        let client = reqwest::Client::new();
        let url = self.to_upload_link_url();
        let resp = client
            .get(&url)
            .header("Authorization", &self.token)
            .query(&[("path", self.repo_path.as_str())])
            .header("User-Agent", "PayslipDownloader")
            .header("Accept", "application/json, text/javascript, */*; q=0.01")
            .send()
            .await
            .expect("Seafile auth is fine");

        let mut link = resp.text().await.expect("Upload link");
        link.pop();
        link.remove(0);
        self.upload_link = link;
    }
}

#[derive(Clone, Debug, Serialize)]
#[serde(rename_all = "PascalCase")]
struct DocumentStateChangeValue {
    is_read: bool,
    workflow_step_id: String,
}

#[derive(Clone, Debug, Serialize)]
#[serde(rename_all = "PascalCase")]
struct DocumentStateChange {
    value: DocumentStateChangeValue,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
struct Document {
    id: String,
    name: String,
    workflow_step_id: String,
    domain: i32,
    label: String,
    description: Option<String>,
    application_id: i32,
    blob_id: Option<String>,
    file_name: Option<String>,
    file_size: i64,
    entry_type: i32,
    created: String,
    updated: Option<String>,
    created_by_member_id: Option<String>,
    updated_by_member_id: Option<String>,
    created_by_organization_id: Option<String>,
    created_by_member_name: Option<String>,
    created_by_organization_name: Option<String>,
    period_from: Option<String>,
    period_to: Option<String>,
    has_been_read: bool,
    has_been_deleted: bool,
    has_been_locked: bool,
    has_been_archived: bool,
}

impl Document {
    fn change_state(&self, auth: &OneclickAuth, is_read: bool) -> bool {
        let body = DocumentStateChange {
            value: DocumentStateChangeValue {
                is_read: is_read,
                workflow_step_id: self.workflow_step_id.clone(),
            },
        };

        let client = reqwest::blocking::Client::new();
        let url = auth.to_document_put_state(&self.id);
        let client_referer = auth.client_url.clone();
        let client_origin = auth.client_url.clone();
        let resp = client
            .put(&url)
            .json(&body)
            .header(
                "User-Agent",
                "Mozilla/5.0 (X11; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0",
            )
            .header("Accept", "application/json, text/javascript, */*; q=0.01")
            .header("Referer", client_referer.as_str())
            .bearer_auth(&auth.token)
            .header("Origin", client_origin.as_str())
            .send()
            .expect("State change is fine");

        resp.status().is_success()
    }

    fn is_folder(&self) -> bool {
        self.entry_type == 1
    }
}

#[derive(Clone, Debug, Deserialize)]
//#[serde(rename_all = "PascalCase")]
struct DocumentList {
    items: Vec<Document>,
    //next_page_link: Option<String>,
    //count: i32
}
#[derive(Clone, Debug, Deserialize)]
#[serde(rename_all = "PascalCase")]
struct DocumentDownloadResponse {
    url: String,
    //blob_name: String,
    //container_name: String,
    //folder_name: Option<String>,
    //ready: String
}

/*#[derive(Clone, Debug, Deserialize)]
struct DocumentUploadResponse {
    name: String,
    id: String,
    size: i64
}*/

#[derive(Clone, Debug, Serialize, Deserialize)]
struct DocumentProcessingStatus {
    doc: Document,
    file_name: String,
    status: bool,
}

impl DocumentProcessingStatus {
    fn to_string(&self) -> String {
        let mut s = format!("Processing of file {} ", self.doc.id);
        if self.status {
            s.push_str("was successful.");
            s.push_str(" New file name: ");
            s.push_str(self.file_name.as_str());
        } else {
            s.push_str("failed.");
        }

        s.clone()
    }
}

fn generate_file_name(file_name: &String) -> String {
    let mut t = file_name.as_str().split("_");
    let items = t.clone().count();
    if items != 8 {
        file_name.clone()
    } else {
        let date = t.nth(5).expect("Date");
        let year = date.get(0..4).unwrap();
        let month = date.get(4..6).unwrap();
        let day = date.get(6..8).unwrap();
        let new_file_name = format!("{}_{}_{}_{}", year, month, day, file_name);
        new_file_name
    }
}

fn process_document(
    auth: OneclickAuth,
    seafile_api: SeafileApi,
    tx: Sender<DocumentProcessingStatus>,
    doc: Document,
) {
    let file_name = doc.file_name.as_ref().unwrap();
    println!("Processing: Id: {}, Name: {}", doc.id, file_name);

    let client_referer = auth.client_url.clone();
    let client_origin = auth.client_url.clone();
    // request download
    let client = reqwest::blocking::Client::new();
    let url = auth.to_document_download(&doc.blob_id.as_ref().unwrap());
    let resp = client
        .get(&url)
        .query(&[
            ("contentDisposition", "attachment"),
            ("fileName", file_name.as_str()),
        ])
        .header(
            "User-Agent",
            "Mozilla/5.0 (X11; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0",
        )
        .header("Accept", "application/json, text/javascript, */*; q=0.01")
        .header("Referer", client_referer.as_str())
        .bearer_auth(&auth.token)
        .header("Origin", client_origin.as_str())
        .send()
        .expect("Download request is fine");

    let resp_download_document: DocumentDownloadResponse =
        resp.json().expect("Valid response document");
    let resp_download = client
        .get(&resp_download_document.url)
        .header(
            "User-Agent",
            "Mozilla/5.0 (X11; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0",
        )
        .send()
        .expect("Valid document");
    // upload document
    let data = resp_download.bytes().expect("Bytes").to_vec();
    let file_part = reqwest::blocking::multipart::Part::bytes(data);
    let new_file_name = generate_file_name(&file_name.clone());
    let file_part = file_part.file_name(new_file_name.clone());
    let repo_path = seafile_api.repo_path.clone();
    let repo_path = repo_path.as_str();
    let form = reqwest::blocking::multipart::Form::new()
        .text("replace", "1")
        .text("parent_dir", repo_path.to_owned())
        .part("file", file_part);
    let resp_upload = client
        .post(&seafile_api.upload_link)
        .header("Authorization", &seafile_api.token)
        .header(
            "User-Agent",
            "Mozilla/5.0 (X11; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0",
        )
        .query(&[("ret-json", "1")])
        .multipart(form)
        .send()
        .expect("Valid upload");

    let resp_status = resp_upload.status().is_success();

    if resp_status {
        //let _result: Vec<DocumentUploadResponse> = resp_upload.json().expect("Valid upload response");
        // mark document as read
        doc.change_state(&auth, true);
    }

    let dps = DocumentProcessingStatus {
        doc: doc,
        file_name: new_file_name.clone(),
        status: resp_status,
    };
    tx.send(dps).expect("Sending works!");
}

fn send_mail(processed: &Vec<DocumentProcessingStatus>) {
    // Build body
    let mut body = String::from("Hi!\n\nWe've downloaded following payroll information:\n\n");
    for m in processed {
        let t = format!(
            "    - Id: {}, Filename: {}",
            m.doc.id.clone(),
            m.file_name.clone()
        );
        body.push_str(t.as_str());
        body.push_str("\n\n");
    }

    let email = Message::builder()
        .from(env::var("MAIL_FROM").unwrap().parse().unwrap())
        .reply_to(env::var("MAIL_REPLYTO").unwrap().parse().unwrap())
        .to(env::var("MAIL_TO").unwrap().parse().unwrap())
        .subject("New payrolls downloaded")
        .body(body)
        .unwrap();

    let creds = Credentials::new(
        env::var("MAIL_USERNAME").unwrap(),
        env::var("MAIL_PASSWORD").unwrap(),
    );

    // Open a remote connection to mail server
    let relay = env::var("MAIL_HOST").unwrap();
    let mailer = SmtpTransport::relay(relay.as_str())
        .unwrap()
        .credentials(creds)
        .build();

    // Send the email
    match mailer.send(&email) {
        Ok(_) => println!("Email sent successfully!"),
        Err(e) => panic!("Could not send email: {:?}", e),
    }
}

fn process_document_list(auth: &OneclickAuth, resp: Response, seafile_api: &SeafileApi) {
    println!("Document request was: {}", resp.status());
    let resp_json_result: Result<DocumentList, Error> = resp.json();
    let resp_json: DocumentList = resp_json_result.expect("Valid document list!");
    let n_workers = 4;
    let pool = threadpool::ThreadPool::new(n_workers);
    let item_count = resp_json.items.len();
    println!("Retrieved {} documents.", item_count);

    if item_count > 0 {
        let (tx, rx) = channel();
        for d in resp_json.items {
            let tx = tx.clone();
            let auth = auth.clone();
            let seafile_api = seafile_api.clone();
            // Validate the document itself.
            pool.execute(move || {
                if d.is_folder() {
                    document_thread(&auth, &seafile_api, Some(&d.id));
                } else {
                    process_document(auth, seafile_api, tx, d);
                }
            });
        }

        let mut process_list: Vec<DocumentProcessingStatus> =
            Vec::<DocumentProcessingStatus>::new();

        let mut iter = rx.iter();
        for _i in 0..item_count {
            let msg = iter.next().unwrap();
            println!("{}", msg.to_string());
            process_list.push(msg);
        }

        send_mail(&process_list);
    }
    println!("Done...");
}

fn document_thread(auth: &OneclickAuth, seafile_api: &SeafileApi, folder_id: Option<&str>) {
    let client = reqwest::blocking::Client::new();
    println!("Next run...");
    let options = [
        ("applicationId", "3"),
        ("$inlinecount", "allpages"),
        ("$top", "200"),
        ("$orderby", "Created desc"),
        ("$filter", "HasBeenRead eq false"),
    ];
    let mut options = options.to_vec();
    if let Some(folder_id_str) = folder_id {
        options.insert(0,
            ("folderId", &folder_id_str)
        );
    }
    let resp = client
        .get(auth.to_documents())
        .query(&options)
        .header(
            "User-Agent",
            "Mozilla/5.0 (X11; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0",
        )
        .header("Accept", "application/json, text/javascript, */*; q=0.01")
        .header("Referer", auth.client_url.clone())
        .bearer_auth(&auth.token)
        .header("Origin", auth.client_url.clone())
        .send();
    match resp {
        Ok(e) => process_document_list(&auth, e, &seafile_api),
        Err(e) => println!("Document request failed: {}", e.to_string()),
    }
    println!("Job done");
}

async fn execute(auth: &OneclickAuth, auth_seafile: &SeafileApi) {
    let auth = auth.clone();
    let auth_seafile = auth_seafile.clone();
    let dt = thread::spawn(move || {
        document_thread(&auth, &auth_seafile, None);
    });
    dt.join().ok();
}

#[tokio::main]
async fn main() {
    dotenv::dotenv().ok();
    let _guard: Option<ClientInitGuard> = match env::var("SENTRY_DSN") {
        Ok(dsn) => {
            println!("Initialize trace / sentry via: {}", dsn);
            let x = sentry::init((
                dsn,
                sentry::ClientOptions {
                    release: sentry::release_name!(),
                    ..Default::default()
                },
            ));
            Some(x)
        }
        Err(_) => None,
    };

    // Obtain a token to service portal itself
    let auth = OneclickAuth {
        base_url: env::var("ONECLICK_BASEURL").unwrap(),
        client_url: env::var("ONECLICK_CLIENT_URL").unwrap(),
        client_number: env::var("ONECLICK_CLIENT").unwrap(),
        username: env::var("ONECLICK_USERNAME").unwrap(),
        password: env::var("ONECLICK_PASSWORD").unwrap(),
        organization: String::from(""),
        parent_organization: String::from(""),
        token: String::from(""),
    };
    let auth_seafile = SeafileApi {
        base_url: env::var("SEAFILE_BASEURL").unwrap(),
        password: env::var("SEAFILE_PASSWORD").unwrap(),
        token: String::from(""),
        repo_path: env::var("SEAFILE_PATH").unwrap(),
        upload_link: String::from(""),
    };

    loop {
        let mut auth_share = auth.clone();
        // Obtain a token to service portal itself
        auth_share.obtain_token().await;
        auth_share.get_user_info().await;

        // Obtain token to Seafile.
        let mut auth_share_seafile = auth_seafile.clone();
        auth_share_seafile.auth().await;
        auth_share_seafile.get_upload_link().await;

        // Check and download files.
        execute(&auth_share, &auth_share_seafile).await;
        thread::sleep(time::Duration::from_secs(604800))
    }
}
